import csv
from pathlib import Path

import click
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sb
from ncempy.io import dm
from scipy import ndimage as ndi
from skimage import data, filters, io
from skimage.feature import blob_dog, blob_doh, canny, peak_local_max
from skimage.measure import regionprops
from skimage.morphology import label, watershed
from skimage.restoration import denoise_bilateral
from tqdm import tqdm


def threshold(img):
    """Uses Otsu thresholding to reduce background noise.
    Returns binarized image"""
    thresh = filters.threshold_otsu(img)
    return img > thresh


def rolling_ball(img):
    pass


def watershed_segmentation(img):
    distance = ndi.distance_transform_edt(img)
    local_max = peak_local_max(distance, indices=False)
    plt.plot(local_max)
    plt.show()
    markers = ndi.label(local_max)[0]
    return watershed(-distance, markers, compactness=0.01)


def denoising(img):
    return denoise_bilateral(img, multichannel=False)


def edge_detection(img):
    edges = filters.sobel(img)
    return edges


def get_pixel_ratio(img, img_path):
    """
    Tries to read file metadata from dm file. If normal .tif images are provided
    instead, prompts user for the nm/pixel ratio, which can be found using the
    measurement tool in ImageJ or similar. For example, a scale bar of 100nm corresponds to a nm/pixel ratio of
    0.204.
    """
    try:
        try:
            metadata = dm.dmReader(f"{str(img_path)[:-4]}.dm4")
        except FileNotFoundError:
            metadata = dm.dmReader(f"{str(img_path)[:-4]}.dm3")
        ratio = metadata["pixelSize"][0]
    except FileNotFoundError:
        io.imshow(img)
        plt.show()
        ratio = click.prompt(
            f"dm4 not found. Enter the nm/pixel ratio in {img_path.stem}:",
            type=float,
            default=0.204,
        )
        plt.close()

    return ratio


def blob_detection(img):
    """Finds maximas in the determinant of the Hessian of the image,
    which corresponds to particles"""
    return blob_doh(img, min_sigma=26, max_sigma=120, num_sigma=25)


def plot_blobs(img, blobs, outpath):
    """Saves masking of blob boundaries on top of original image
    in results subdirectory"""
    fig, ax = plt.subplots()
    ax.set_axis_off()
    plt.tight_layout()
    ax.imshow(img)
    for blob in blobs:
        y, x, r = blob
        c = plt.Circle((x, y), r, color="r", linewidth=2, fill=False)
        ax.add_patch(c)
    plt.savefig(f"{outpath}_blobs")


def particle_diameters(blobs, ratio):
    diameters = []
    for blob in blobs:
        diameters.append(blob[-1] * 2 * ratio)
    return diameters


def _flatten(nested):
    flat = []
    for lst in nested:
        for item in lst:
            flat.append(item)
    return flat


def plot_kde(result):
    """Plots histogram of particles sizes,
    along with kernel density estimate of distribution"""
    data = np.loadtxt(result)
    plot = sb.distplot(data, axlabel="Diameter (nm)", hist=False)
    plot.figure.savefig(Path(result).with_suffix(".png"), dpi=1200)


def summary_stats(result):
    # TODO: add test to see if analysis passes ISO 13322-1
    print(
        f"""Average: {sum(result)/len(result)} \nMinimum: {min(result)} 
        \nMaximum: {max(result)} """
    )


@click.command()
@click.argument(
    "dir", type=click.Path(exists=True)
)
def main(dir):
    """
    Takes a folder of TEM images (preferably in dm3 or dm4 format) and calculates
    aggregate statistics for particle sizes. Creates a plot and a .csv for
    manual/custom plotting, if desired. Automatically detects and sizes
    particles, and outputs a visualization after each detection so that the
    scientist can determine whether the prediction is accurate.
    """
    io.use_plugin("matplotlib")
    output = Path(f"{dir}/Results")
    output.mkdir(parents=True, exist_ok=True)

    print("Processing folder...")
    result_name = click.prompt(
        "Enter result filename", type=str, default=output.parents[0].name
    )
    csv_name = Path(output / result_name).with_suffix(".csv")

    for idx, img_path in enumerate(Path(dir).glob("**/*.tif")):
        img = io.imread(img_path, as_gray=True)
        ratio = get_pixel_ratio(img, img_path)
        # crop out scale bar
        print("Processing...")
        img = img[:2230, :]
        cleaned = denoising(img)
        blobs = blob_detection(cleaned)

        outpath = Path(dir) / "Results" / str(idx + 1)
        plot_blobs(img, blobs, outpath)
        plt.show()
        if click.confirm("Is this an acceptable result?"):
            plt.close()
            with open(csv_name, "a") as f:
                np.savetxt(f, particle_diameters(blobs, ratio))

        else:
            pass

    plot_kde(csv_name)
    print("Finished!")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 12 16:39:19 2016

@author: dendrondal
"""

import glob
import os
import re
from pathlib import Path

import click
import matplotlib.pyplot as plt
import pandas as pd
from cycler import cycler
from matplotlib import ticker


def tga_plot(cwd, title, filenames=None):
    """
    Plotting utility for TGA. In order to used, TGA must be exported as .xls
    via the analysis software on the TGA computer.
    ------------
    Parameters
    ------------
    filenames: all of the filenames you want to stack on top of each other,
    in quotes, in square brackets. For examle, if you want to compare 90-30
    G3-PAMAM-PLLA, enter ['90-L-G3', '70-L-G3', '50-L-G3', '30-L-G3'].
    If filenames is None, automatically plots everything in directory

    cwd: file location where excel files are located (for example, if
    they're in your flash drive folder, enter 'F:/yourname/' in quotes)
    """

    decomposition = []

    fig, ax = plt.subplots(figsize=(3.0, 2.5))

    if filenames is None:
        filenames = [file for file in Path(cwd).glob("*.txt")]

    for name in filenames:
        df = pd.read_excel(f"{cwd}/{name}", skiprows=20)
        TG_ = [100]
        # converts raw mass measurements into % decomposition
        for j in range(1, len(df["ug"])):
            TG_.append((df["ug"][j] / df["ug"][0]) * 100)
        df["TG(Percent)"] = TG_
        # list of decomposition temps is made to use in plot labels or output.
        decomposition.append(
            df.loc[df["TG(Percent)"].apply(lambda x: int(x)) == 95, "Cel"].mean()
        )

        # Previously used plotting parameters are greyed out. May still be wanted.
        # plt.title(title, family='arial', weight='bold')
        plt.xlabel("Temperature", family="arial", weight="bold", size=10)
        # plt.xticks(np.arange(df['Cel'].min(), df['Cel'].max(),10),
        #           labels=None)
        plt.ylabel("%TG", family="arial", weight="bold", size=10)
        # plt.yticks(np.arange(df['TG(Percent)'].min(), df['TG(Percent)'].max(),1),
        #           labels=None)
        # fig.patch.set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        # ax.axis('off')
        plt.plot(df["Cel"], TG_, label=f"{name[:-4].upper()}", linewidth=2)
        plt.tight_layout()
        plt.legend()
        plt.savefig(os.path.join(cwd, f"{title}.png"), dpi=1440)

    return decomposition


@click.command()
@click.option(
    "--path",
    prompt="Folder containing TGA data",
    type=click.Path(exists=True),
    default=os.getcwd(),
)
@click.option("--title", prompt="Plot title")
def main(path, title):
    return tga_plot(path, title)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 12 16:39:19 2016

@author: dendrondal
"""
import glob
import os
import re
from pathlib import Path

import click
import matplotlib.pyplot as plt
import pandas as pd
from cycler import cycler
from matplotlib import ticker


def csv_extraction(path, save=False):
    """Takes in .txt file output by TA software and parses into individual
    cyles. If save boolean is true, saves .csv version of file
    with same name as TA .txt output."""
    with open(path, encoding="utf-16") as f:
        for i, line in enumerate(f):
            if "OrgFile" in line:
                # Some files are output with different starting lines.
                start = i + 1
                break
            else:
                start = 68
                # default starting line for output data
        df = pd.read_csv(
            path, delim_whitespace=True, skiprows=start, header=None, encoding="utf-16"
        )

    df.columns = [
        "time (min)",
        "temperature (C)",
        "heat flow (mW)",
        "heat capacity (mJ/C)",
        "N2 flow",
    ]

    num = 0

    def cycle(x):
        nonlocal num
        if x["time (min)"] == -2:
            # for some reason, this denotes a switch between heating/cooling
            num = x["temperature (C)"]
        return num

    df["cycle"] = df.apply(cycle, axis=1)
    if save:
        df.to_csv(path.with_suffix(".csv"))

    return df


def save_multiple_csvs(cwd):
    os.chdir(cwd)
    for file in glob.glob("*.txt"):
        csv_extraction(file, save=True)


def _dsc_plot_setup(legend):
    fig, ax = plt.subplots(figsize=(3.5, 3.0))
    ax.set_prop_cycle(cycler("color", ["y", "g", "b", "r"]))
    # Plot formatting goes here
    plt.xlabel("Temperature ($^o$C)", family="arial", weight="bold", size=8)
    plt.ylabel("Heat Flow (mW)", family="arial", weight="bold", size=8)
    minor_locator = ticker.AutoMinorLocator()
    ax.xaxis.set_minor_locator(minor_locator)
    ax.xaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.tick_params(axis="both", which="major", labelsize=6)
    # custom y-axis, if needed:
    # ax.set_ylim([-1, -0.6])
    plt.tight_layout()

    return fig, ax


def dsc_plotting(cwd, title, cycle=2, filenames=None, legend=False):
    """Creates stacked plot of all DSC data from multiple .csvs
    ------------
    Parameters
    ------------
    cwd: file location where csv files are located (for example, if
    they're in your flash drive folder, enter 'F:/yourname/datafolder' in quotes)

    title: title of saved .png file

    filenames: all of the filenames you want to stack on top of each other,
    in quotes, in square brackets. For examle, if you want to compare 90-30
    G3-PAMAM-PLLA, enter ['90-L-G3', '70-L-G3', '50-L-G3', '30-L-G3'].
    If filenames is None, automatically plots over all .txt files in cwd.

    cyle: cycle number to be plotted.

    legend: whether or not to include legend object with plot.

    """
    cycle = float(cycle)
    fig, ax = _dsc_plot_setup(legend)

    if filenames is None:
        filenames = [file for file in Path(cwd).glob("*.txt")]
    elif type(filenames) is str:
        filenames = list(filenames)
    else:
        raise TypeError("filenames must be list of strings or None.")

    for i, name in enumerate(filenames):
        print(f"Processing {name.stem}...")
        df = csv_extraction(name)
        df = normalize(df, cycle)
        # This operation introduces offset between the graphs
        df["normalized"] = df["normalized"].apply(lambda x: x - i * 1 / len(filenames))
        if len(df.loc[df["cycle"] == cycle, "temperature (C)"][:-1]) == 0:
            raise Exception(
                "Invalid cycle specified. Please use enter a cycle between {} and {}".format(
                    df["cycle"].min(), df["cycle"].max()
                )
            )
        label = str(name.stem).split("_")[1]
        ax.plot(
            df.loc[df["cycle"] == cycle, "temperature (C)"][:-1][1:],
            df.loc[df["cycle"] == cycle, "normalized"][:-1][1:],
            label=label,
            linewidth=1.25,
        )
    if legend:
        ax.legend(prop={"size": 6}, frameon=False)

    plt.savefig(Path(cwd) / f"{title}.png", dpi=300)


def normalize(df, cycle):
    min_mW = df.groupby([df["cycle"] == cycle]).min()["heat flow (mW)"][1]
    max_mW = df.groupby([df["cycle"] == cycle]).max()["heat flow (mW)"][1]
    if min_mW == max_mW:
        cycle += 1
        print(f"Dummy cycle specified. Trying cycle {cycle} instead")
    # Getting rid of settingwithcopy warning, since it doesn't matter here
    pd.options.mode.chained_assignment = None
    df["normalized"] = df.loc[:, "heat flow (mW)"].apply(
        lambda x: (x - max_mW) / (max_mW - min_mW)
    )
    return df


def cycle_reproducibility(path, odd_heat_cycles=False):
    """Tests to see if the DSC curves look similar between cycles."""
    os.chdir(path)
    filenames = [file for file in glob.glob("*.txt")]
    for polymer in filenames:
        title = polymer.split("_")[-1][:-4].upper()
        plt.title(title)
        df = csv_extraction(polymer)
        heating = df.loc[
            (df["cycle"] % 2 == 0 + odd_heat_cycles)
            & (df["cycle"] != 0 + odd_heat_cycles)
            & (df["cycle"] != 12 - odd_heat_cycles)
            & (df["time (min)"] > 0)
        ]
        # Following loop finds best way to split up plots
        for coord in range(1, 4, -1):
            if len(filenames) % coord == 0:
                fig, axes = plt.subplots(len(filenames) / coord, coord)

        for cycle, ax in zip(heating["cycle"].unique(), axes.flatten()):
            heating.loc[df["cycle"] == cycle].plot(
                x="temperature (C)", y="heat flow (mW)", ax=ax
            )

        plt.savefig(os.path.join(path, f"{title}.png"))


@click.command()
@click.option(
    "--path",
    prompt="Folder containing DSC data",
    help="Folder containing DSC data",
    type=click.Path(exists=True),
    default="/home/dal/Documents/jsdw_191217_pisa-dsc",
)
@click.option("--title", prompt="Plot title")
@click.option("--cycle", prompt="Cycle number to plot")
def main(path, title, cycle):
    leg = click.confirm("Include legend?")
    return dsc_plotting(path, title, cycle, legend=leg)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 22 18:08:18 2016

@author: dendrondal
"""
import click
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import style

style.use("bmh")


@click.command()
@click.option(
    "--path",
    prompt="Full path to .csv file",
    help="Full path to .csv file",
    type=click.Path(exists=True),
)
@click.option(
    "--wv_range",
    prompt="Range of wavelengths (i.e. max - min)",
    help="Range of wavelengths (i.e. max - min)",
)
@click.option(
    "--min_conc",
    prompt="Minimum concentration, in mg/mL",
    help="Minimum concentration, in mg/mL",
)
@click.option(
    "--max_conc",
    prompt="Maximum concentration, in mg/mL",
    help="Maximum concentration, in mg/mL",
)
@click.option(
    "--step",
    prompt="Total number of samples in csv",
    help="Total number of samples in csv",
)
@click.option(
    "--vb1",
    prompt="Lambda max/first vibronic band",
    help="Lambda max/first vibronic band",
)
@click.option(
    "--vb3",
    prompt="Third vibronic band",
    required=False,
    help="If you are comparing the first and third vibronic band for\
              a dye, (i.e. pyrene), enter it here. Otherwise, just press enter",
)
def main(path, wv_range, min_conc, max_conc, step, vb1, vb3=None):
    """
    Graphs .csv output from Bruker UV-Vis software. Outputs stacked UV-vis
    spectra and wavelength (or wavelength raio, depending on dye) vs.
    log(concentration) spectra as .png files in the same directory as .csv file.
    """
    # creates and saves overlapping UV-vis spectra for each sample
    df = pd.read_csv(path, header=None, skiprows=2)
    even_cols = [col for col in range(2, len(df.columns)) if col % 2 == 0]
    # removes all even columns past index 0, corresponding to wavelength repeats
    df = df.drop(even_cols, 1)
    df = df[: wv_range + 1]
    df = df.convert_objects(convert_numeric=True)
    conc = np.linspace(min_conc, max_conc, step)
    plt.figure(figsize=(10, 7))
    for i in range(1, len(conc) + 1, 2):
        plt.plot(df[0].values, df[i].values, label=i)
    abs_vb1 = df[df[0] == vb1].values
    plt.savefig(f"{path[:-4]}_stacked_UV-vis.png")
    plt.show()
    # creates CAC graphs with lambda max or 1st and 3rd vibronics
    log_conc = [np.log(value) for value in conc]
    plt.figure(figsize=(10, 7))
    plt.xlabel("ln(Concentration (mg/mL))")
    if not vb3:
        plt.ylabel(f"Absorbance intensity: {vb1} nm")
        plt.plot(log_conc, abs_vb1[0, 1:], marker="*")
    else:
        df["ratio"] = df[df[0] == vb1] / df[df[0] == vb3]
        plt.ylabel(f"I{vb1}/I{vb3}")
        plt.plot(log_conc, df["ratio"].values, marker="*")
    plt.savefig(f"{path[:-4]}_cac.png")
    plt.show()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['instrumentools']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'matplotlib>=3.1.2,<4.0.0',
 'ncempy>=1.5.0,<2.0.0',
 'numpy>=1.17.4,<2.0.0',
 'pandas>=1.0.0,<2.0.0',
 'scikit-image>=0.16.2,<0.17.0',
 'seaborn>=0.9.0,<0.10.0',
 'tqdm>=4.40.2,<5.0.0',
 'xlrd>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['cac_analysis = instrumentools.CAC:main',
                     'dsc_analysis = instrumentools.DSC:main',
                     'tem_analysis = instrumentools.TEM:main',
                     'tga_analysis = instrumentools.TGA:main']}

setup_kwargs = {
    'name': 'instrumentools',
    'version': '0.1.3',
    'description': 'Data processing and plotting scripts for chemistry workflows',
    'long_description': None,
    'author': 'Dal Williams',
    'author_email': 'dendrondal@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dendrondal/instrumentools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
